#! /usr/bin/python

#
# SMTPServer.py 0.9
#                                                                      
# Copyright (C) 2001,2002 Gerson Kurz (not@p-nand-q.com)
# BSD-Licensed. This is free software, enjoy!

"""SHICKS! SMTP Server for Python

This server implements a RFC 821 compliant SMTP server with a MySQL-based
mailbox storage. The RFC can be found online at

    http://www.ietf.org/rfc/rfc821.txt

The comments for the commands below are taken from that document. The
database format is discussed in README.HTM. This module is designed to
be reusable in other python projects.
"""

import sys, socket, SocketServer, traceback, cStringIO, time, hmac
import smtplib, binascii, os, ShicksTrace, fnmatch, base64, thread, rotor
import mxlookup

from SMTPStatusMessages import status_msgs
from SMTPUtilities import *
from ShicksConfig import GetConfigSettings
from ShicksSpamCheck import * # spamcheck handling
import ShicksUser # user management
import ShicksMessageList # message list handling

CURRENT_VERSION = "0.9"

# If this flag is set, SMTP-AUTH must be used
REQUIRE_SMTP_AUTH = int(GetConfigSettings("RequireSmtpAuth","0"))

# If this flag is set, mails to <unknown>@local-domain will be forwarded to the administrative account
CATCH_ALL = int(GetConfigSettings("CatchAll","0"))

# This is the relay SMTP server used for mail sent to remote recipients
RELAY_SMTP_SERVER_NAME = GetConfigSettings("SmtpRelayServer","")
USE_RELAY_SMTP_SERVER = int(GetConfigSettings("UseRelaySmtpServer","0"))

default_storage = "/usr/share/shicks/mail"
if os.name == "nt":
    default_storage = "c:\\shicks\\mail"
LOCAL_STORAGE = GetConfigSettings("LocalStorage",default_storage)

ALLOW_UNKNOWN_RECEIVER_DEFAULT = int(GetConfigSettings("AllowUnknownReceiverDefault","1"))
ALLOW_UNKNOWN_SENDER_DEFAULT = int(GetConfigSettings("AllowUnknownReceiverDefault","1"))
ADMIN_ACCOUNT = GetConfigSettings("AdminAccount","")

# mails that use this domain
DEFAULT_LOCAL_DOMAIN = GetConfigSettings("LocalDomain","")

# will be forwarded to this account
UNKNOWN_ACCOUNT = GetConfigSettings("UnknownLocalAccount","")

# Strict test for relaying. If enabled (by default), either the sender or ALL receivers must
# have a known local address. If disabled, either the sender or ONE OF the receivers must be a
# known local address
STRICT_RELAY_TEST = int(GetConfigSettings("StrictRelayTest","1"))

IP_ADDRESS_RANGE = decodeAddressRangeString(GetConfigSettings("LocalIpRange",""))
DENY_HOSTS_LIST = decodeAddressRangeString(GetConfigSettings("DenyHostsList",""))
BLACKLIST_DOMAINS = decodeAddressRangeString(GetConfigSettings("BlacklistDomains",";".join(BLACKLIST_DOMAINS)))
BLACKLIST_IGNORE = decodeAddressRangeString(GetConfigSettings("BlacklistIgnore",";".join(BLACKLIST_IGNORE)))

try:
    os.makedirs(LOCAL_STORAGE)
except:
    pass
UNIQUE_ID_COUNTER = 1L

class SMTPServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
    allow_reuse_address = 1    # Seems to make sense in testing environment
    def server_bind(self):
        """Override server_bind to store the server name."""
        SocketServer.TCPServer.server_bind(self)
        host, port = self.socket.getsockname()
        self.server_name = socket.getfqdn(host)
        self.server_port = port

class SMTPRequestHandler(SocketServer.StreamRequestHandler):

    def SendResponse(self,msg,args=None):
        # messages that indicate abuse => stop communication ASAP
        if msg in ('ERR_ACC_DENIED',
                   'ERR_ACC_DENIED_BY_BLACKLIST',
                   'ERR_ACC_DENIED_BY_RULE',
                   'ERR_ACC_DENIED_UNKNOWN_USER'):
            self.terminated = 1
            
        msg = self.status_messages[msg]
        if args is not None:
            msg = msg % args
        print "WRITE:" + msg.rstrip()
        self.wfile.write(msg)
        self.wfile.flush()      
            
    def ResetMessage(self):        
        self.message = { 'FROM':None, 'TO':[], 'MSG':[] }        

    def handle(self):
        # SMTP mailbox state variables        
        self.terminated = 0
        self.receiving_data = 0
        self.ResetMessage()

        self.status_messages = status_msgs

        # AUTH handling
        self.AuthSecret = "%d.%d.%s@%s.%s" % (os.getpid(),thread.get_ident(),str(time.clock()),socket.gethostname(),str(id(self)))
        self.AuthPending = None
        self.AuthDone = 0

        # supported commands: all required and optional commands as mentioned in said RFC
        cmdmap = { 'HELO' : self.SendGreeting,
                   'QUIT' : self.CloseConnection,
                   'MAIL' : self.MailCommand,
                   'RCPT' : self.ReceiveTo,
                   'NOOP' : self.NoOperation,
                   'RSET' : self.ResetOperation,
                   'VRFY' : self.VerifyAddress,
                   'EXPN' : self.ExpandList,
                   'HELP' : self.HelpMessage,
                   'EHLO' : self.SendExtendedGreeting,
                   'AUTH' : self.Auth,
                   'DATA' : self.StartReceivingData }

        print "BEGIN of SMTP conversation with " + str(self.client_address)

        # determine if this is a local connection. if it is, then
        # some restrictions are lowered.
        ip_address = self.client_address[0]
        self.sender_host = self.ip_address_string = str(ip_address)
        self.is_local_connection = 0
        global IP_ADDRESS_RANGE
        if IP_ADDRESS_RANGE is not None:
            if addressInRange(ip_address,IP_ADDRESS_RANGE):
                self.is_local_connection = 1
                print "Connection is local, skipping relay checks."

        if not self.is_local_connection:
            
            # check if this comes from a known spammer; if so, reject conversation
            blacklist_host = checkRelayBlacklist(ip_address)
            if blacklist_host is not None:
                print "Host '%s' is blacklisted by %s, denying access" % (ip_address, blacklist_host)
                return self.SendResponse('ERR_ACC_DENIED_BY_BLACKLIST',(ip_address, blacklist_host))

            # check if this mail is on the local deny list
            if addressInRange(ip_address, DENY_HOSTS_LIST):
                print "Host '%s' is on the local deny-hosts-list, denying access" % (ip_address)
                return self.SendResponse('ERR_ACC_DENIED_BY_RULE', (ip_address))
        
        self.SendResponse("OK_WELCOME",(CURRENT_VERSION))
        
        elcount = 0

        while not self.terminated:        
            self.inputline = self.rfile.readline()
            line = self.inputline.strip()

            if self.receiving_data:
                if self.inputline == '.\r\n':
                    self.SendMessage()
                else:                    
                    self.message['MSG'].append(self.inputline)
                continue
            elif line == '':
                print "Got empty line %d" % elcount
                elcount += 1
                if elcount == 10:
                    break
                continue

            elcount = 0
            print "READ:" + line.strip()

            if self.AuthPending:
                self.AuthResponse(line)
                continue

            request = splitMailHeader(line)
            # smtp is case-insensitive
            request[0] = request[0].upper()            
            if cmdmap.has_key(request[0]):
                cmdmap[request[0]](request)
            else:
                self.SendResponse('ERR_CMD_UNKNOWN',(request[0]))
        
        print "END of SMTP conversation with " + str(self.client_address)

    def VerifyAddress(self,request):
        """VERIFY (VRFY)

        This command asks the receiver to confirm that the argument
        identifies a user.  If it is a user name, the full name of
        the user (if known) and the fully specified mailbox are
        returned.

        This command has no effect on any of the reverse-path
        buffer, the forward-path buffer, or the mail data buffer."""

        address = request[1]
        name = self.GetExpandedName(address)
        if name:
            self.SendResponse('OK_VERIFY',name)
        else:
            self.SendResponse('ERR_USR_UNKNOWN',name)

    def ExpandList(self,request):
        """EXPAND (EXPN)

        This command asks the receiver to confirm that the argument
        identifies a mailing list, and if so, to return the
        membership of that list.  The full name of the users (if
        known) and the fully specified mailboxes are returned in a
        multiline reply.

        This command has no effect on any of the reverse-path
        buffer, the forward-path buffer, or the mail data buffer."""

        # todo: add mailing list support        
        self.SendResponse('ERR_UNKNOWN_MBOX')
        
    def HelpMessage(self,request):
        """HELP (HELP)

        This command causes the receiver to send helpful information
        to the sender of the HELP command.  The command may take an
        argument (e.g., any command name) and return more specific
        information as a response.

        This command has no effect on any of the reverse-path
        buffer, the forward-path buffer, or the mail data buffer."""

        self.SendResponse('OK_HELP')

    def ResetOperation(self,request):
        """RESET (RSET)

        This command specifies that the current mail transaction is
        to be aborted.  Any stored sender, recipients, and mail data
        must be discarded, and all buffers and state tables cleared.
        The receiver must send an OK reply."""
        self.ResetMessage()

        self.SendResponse('OK_RESET')

    def SendGreeting(self,request):
        """HELLO (HELO)

        This command is used to identify the sender-SMTP to the
        receiver-SMTP.  The argument field contains the host name of
        the sender-SMTP.

        The receiver-SMTP identifies itself to the sender-SMTP in
        the connection greeting reply, and in the response to this
        command.

        This command and an OK reply to it confirm that both the
        sender-SMTP and the receiver-SMTP are in the initial state,
        that is, there is no transaction in progress and all state
        tables and buffers are cleared."""

        if len(request) > 1:
            self.sender_host = request[1]
            ip_address = getIpAddress(self.sender_host)
            try:
                ip_address = socket.gethostbyname(self.sender_host)
            except:
                return self.SendResponse('ERR_ACC_DENIED_INVALID_DOMAIN',(self.sender_host))

            if ip_address <> self.ip_address_string:
                # I'm not sure if that is a fatal condition or not, so I'll just warn here
                print "Warning, host '%s' resolves to %s, but mail comes from %s." % (self.sender_host, ip_address, self.ip_address_string)
            
        else:
            # send access denied, because user *must* send domain
            return self.SendResponse('ERR_ACC_DENIED_NO_DOMAIN')
            
        self.SendResponse('OK_GREETING',(self.sender_host,self.ip_address_string))

    def Auth(self,request):
        # see http://www.cis.ohio-state.edu/cgi-bin/rfc/rfc2554.html for details

        print "requested auth method=%s" % request[1]
        if request[1] == 'CRAM-MD5':
            # cram-md5: used e.g. by pythons' smtplib
            self.AuthPending = request[1]
            return self.SendResponse('OK_SEND_AUTH_REQUIRED',base64.encodestring(self.AuthSecret).strip())

        if request[1] == 'LOGIN':
            # used by mozilla
            if len(request) == 3:
                username = base64.decodestring(request[2])

                user = ShicksUser.GetUserByName(username)
                if not user:
                    print "ERROR, user '%s' is not known" % username
                    return self.SendResponse('ERR_AUTH_FAILED')

                self.AuthPending = (request[1], user)
                print "Have pending AUTH request for user '%s'" % username
                return self.SendResponse('OK_SEND_AUTH_NEXT')
            else:
                # used by outlook/outlook express
                self.AuthPending = ["OUTLOOK", None, None] # username not yet known
                return self.SendResponse('OK_SEND_AUTH_REQUIRED',base64.encodestring("Username:").strip())

        return self.SendResponse('ERR_AUTH_TYPE')

    def IsValidAuthResponse(self, response):
        # find name
        x = response.rfind(" ")
        if x < 0:
            print "ERROR, didn't find user name in auth response"
            return 0

        # find associated password from db
        username = response[:x]
        user = ShicksUser.GetUserByName(username)
        if not user:
            print "ERROR, '%s' is not a valid username" % username
            return 0

        # create expected result
        expected_response = username + " " + hmac.HMAC(user.getPassword(), self.AuthSecret).hexdigest()
        if response != expected_response:
            print "ERROR, didn't match expected response"
            return 0

        print "OK, auth successfull"
        return 1

    def AuthResponse(self,line):
        response = base64.decodestring(line)

        if self.AuthPending == 'CRAM-MD5':          
            
            if self.IsValidAuthResponse(response):
                self.SendResponse('OK_AUTH')
                self.AuthDone = 1
            else:
                self.SendResponse('ERR_AUTH_FAILED')

        elif self.AuthPending[0] == 'OUTLOOK':

            # if this the username ?
            if self.AuthPending[1] is None:

                username = response
                user = ShicksUser.GetUserByName(username)
                if not user:
                    print "ERROR, '%s' is not a valid username" % username
                    return self.SendResponse('ERR_AUTH_FAILED')
                
                self.AuthPending = ["OUTLOOK", user]
                return self.SendResponse('OK_SEND_AUTH_NEXT',base64.encodestring("Password:").strip())

            else: # this is the password
                if self.AuthPending[1].getPassword() == response:
                    self.SendResponse('OK_AUTH')
                    self.AuthDone = 1
                    print "OK, auth successfull"
                else:
                    print "ERROR, invalid password"
                    self.SendResponse('ERR_AUTH_FAILED')

        elif self.AuthPending[0] == 'LOGIN':
            if self.AuthPending[1].getPassword() == response:
                self.SendResponse('OK_AUTH')
                self.AuthDone = 1
                print "OK, auth successfull"
            else:
                print "ERROR, invalid password"
                self.SendResponse('ERR_AUTH_FAILED')

        self.AuthPending = None
        

    def SendExtendedGreeting(self,request):
        # See http://www.faqs.org/rfcs/rfc2821.html for details
        
        if len(request) > 1:
            print "sender is '%s'" % request[1]
            
            self.sender_host = request[1]
            ip_address = getIpAddress(self.sender_host)
            try:
                ip_address = socket.gethostbyname(self.sender_host)
            except:
                return self.SendResponse('ERR_ACC_DENIED_INVALID_DOMAIN',(self.sender_host))

            if ip_address <> self.ip_address_string:
                # I'm not sure if that is a fatal condition or not, so I'll just warn here
                print "Warning, host '%s' resolves to %s, but mail comes from %s." % (self.sender_host, ip_address, self.ip_address_string)
            
        else:
            # send access denied, because user *must* send domain
            return self.SendResponse('ERR_ACC_DENIED_NO_DOMAIN')

        self.SendResponse('OK_EHLO',(CURRENT_VERSION))
        
    def CloseConnection(self,request):
        """QUIT (QUIT)

        This command specifies that the receiver must send an OK
        reply, and then close the transmission channel.

        The receiver should not close the transmission channel until
        it receives and replies to a QUIT command (even if there was
        an error).  The sender should not close the transmission
        channel until it send a QUIT command and receives the reply
        (even if there was an error response to a previous command).
        If the connection is closed prematurely the receiver should
        act as if a RSET command had been received (canceling any
        pending transaction, but not undoing any previously
        completed transaction), the sender should act as if the
        command or transaction in progress had received a temporary
        error (4xx)."""

        self.SendResponse("OK_QUIT")
        self.terminated = 1

    def NoOperation(self,request):
        """NOOP (NOOP)

        This command does not affect any parameters or previously
        entered commands.  It specifies no action other than that
        the receiver send an OK reply.

        This command has no effect on any of the reverse-path
        buffer, the forward-path buffer, or the mail data buffer."""
        self.SendResponse('OK_NOOP')

    def MailCommand(self,request):
        """MAIL (MAIL)

        MAIL FROM:<reverse-path> [OPTIONS] <CRLF>

        This command is used to initiate a mail transaction in which
        the mail data is delivered to one or more mailboxes.  The
        argument field contains a reverse-path.

        The reverse-path consists of an optional list of hosts and
        the sender mailbox.  When the list of hosts is present, it
        is a "reverse" source route and indicates that the mail was
        relayed through each host on the list (the first host in the
        list was the most recent relay).  This list is used as a
        source route to return non-delivery notices to the sender.
        As each relay host adds itself to the beginning of the list,
        it must use its name as known in the IPCE to which it is
        relaying the mail rather than the IPCE from which the mail
        came (if they are different).  In some types of error
        reporting messages (for example, undeliverable mail
        notifications) the reverse-path may be null (see Example 7).

        This command clears the reverse-path buffer, the
        forward-path buffer, and the mail data buffer; and inserts
        the reverse-path information from this command into the
        reverse-path buffer."""

        if REQUIRE_SMTP_AUTH and not self.AuthDone:
            print "ERROR, cannot send mail as long as AUTH not completed."
            return self.SendResponse('ERR_AUTH_REQ')

        if request[1].upper() != 'FROM:':
            return self.SendResponse('ERR_MISS_FROM')

        if not self.IsSenderAllowed(request[2]):
            return self.SendResponse('ERR_ACC_DENIED_UNKNOWN_USER',request[2])

        self.message["FROM"] = request[2]
        self.SendResponse('OK_FROM_ACCPT')

    def ReceiveTo(self,request):
        """RECIPIENT (RCPT)

        This command is used to identify an individual recipient of
        the mail data; multiple recipients are specified by multiple
        use of this command.

        The forward-path consists of an optional list of hosts and a
        required destination mailbox.  When the list of hosts is
        present, it is a source route and indicates that the mail
        must be relayed to the next host on the list.  If the
        receiver-SMTP does not implement the relay function it may
        user the same reply it would for an unknown local user
        (550).

        When mail is relayed, the relay host must remove itself from
        the beginning forward-path and put itself at the beginning
        of the reverse-path.  When mail reaches its ultimate
        destination (the forward-path contains only a destination
        mailbox), the receiver-SMTP inserts it into the destination
        mailbox in accordance with its host mail conventions.

        For example, mail received at relay host A with arguments

              FROM:<USERX@HOSTY.ARPA>
              TO:<@HOSTA.ARPA,@HOSTB.ARPA:USERC@HOSTD.ARPA>

           will be relayed on to host B with arguments

              FROM:<@HOSTA.ARPA:USERX@HOSTY.ARPA>
              TO:<@HOSTB.ARPA:USERC@HOSTD.ARPA>.

        This command causes its forward-path argument to be appended
        to the forward-path buffer."""
    
        if request[1].upper() != 'TO:':
            return self.SendResponse('ERR_MISS_TO')

        if not self.IsReceiverAllowed(request[2]):
            return self.SendResponse('ERR_ACC_DENIED_UNKNOWN_USER',request[2])

        self.message["TO"].append(request[2])
        self.SendResponse('OK_TO_ACCPT')

    def StartReceivingData(self,request):
        """DATA (DATA)

        The receiver treats the lines following the command as mail
        data from the sender.  This command causes the mail data
        from this command to be appended to the mail data buffer.
        The mail data may contain any of the 128 ASCII character
        codes.

        The mail data is terminated by a line containing only a
        period, that is the character sequence "<CRLF>.<CRLF>" (see
        Section 4.5.2 on Transparency).  This is the end of mail
        data indication.

        The end of mail data indication requires that the receiver
        must now process the stored mail transaction information.
        This processing consumes the information in the reverse-path
        buffer, the forward-path buffer, and the mail data buffer,
        and on the completion of this command these buffers are
        cleared.  If the processing is successful the receiver must
        send an OK reply.  If the processing fails completely the
        receiver must send a failure reply.

        When the receiver-SMTP accepts a message either for relaying
        or for final delivery it inserts at the beginning of the
        mail data a time stamp line.  The time stamp line indicates
        the identity of the host that sent the message, and the
        identity of the host that received the message (and is
        inserting this time stamp), and the date and time the
        message was received.  Relayed messages will have multiple
        time stamp lines.

        When the receiver-SMTP makes the "final delivery" of a
        message it inserts at the beginning of the mail data a
        return path line.  The return path line preserves the
        information in the <reverse-path> from the MAIL command.
        Here, final delivery means the message leaves the SMTP
        world.  Normally, this would mean it has been delivered to
        the destination user, but in some cases it may be further
        processed and transmitted by another mail system.

           It is possible for the mailbox in the return path be
           different from the actual sender's mailbox, for example,
           if error responses are to be delivered a special error
           handling mailbox rather than the message senders.

        The preceding two paragraphs imply that the final mail data
        will begin with a  return path line, followed by one or more
        time stamp lines.  These lines will be followed by the mail
        data header and body [2].  See Example 8.

        Special mention is needed of the response and further action
        required when the processing following the end of mail data
        indication is partially successful.  This could arise if
        after accepting several recipients and the mail data, the
        receiver-SMTP finds that the mail data can be successfully
        delivered to some of the recipients, but it cannot be to
        others (for example, due to mailbox space allocation
        problems).  In such a situation, the response to the DATA
        command must be an OK reply.  But, the receiver-SMTP must
        compose and send an "undeliverable mail" notification
        message to the originator of the message.  Either a single
        notification which lists all of the recipients that failed
        to get the message, or separate notification messages must
        be sent for each failed recipient (see Example 7).  All
        undeliverable mail notification messages are sent using the
        MAIL command (even if they result from processing a SEND,
        SOML, or SAML command)."""
        
        if not self.is_local_connection and not self.CanSendMessage():
            self.terminated = 1
            return self.SendResponse('ERR_ACC_DENIED')

        self.receiving_data = 1
        self.SendResponse('OK_TSFR_START')

    def CreateUniqueMessageFilename(self):
        global UNIQUE_ID_COUNTER, LOCAL_STORAGE
        
        # create target directory name
        directory = os.path.join(LOCAL_STORAGE,os.sep.join(map(str,time.localtime())[:3]))

        # make sure directory exist            
        try:
            os.makedirs(directory)
        except:
            pass

        unique_filename = "%d.%s.%d" % (os.getpid(), str(time.clock()), UNIQUE_ID_COUNTER )
        UNIQUE_ID_COUNTER += 1
        return os.path.join(directory,unique_filename)

    def RequiresSpamCheck(self):
        # if the sender is a locally known address, ignore spam check
        # so, if I send a message, it will never be checked for spam ;) 
        address = self.message['FROM']
        if address[0] == '<' or address[-1] == '>':
            address = address[1:-1]

        print "checking sender '%s'" % address

        # if the receiver is a local user
        user = ShicksUser.GetUserByAddress(address)
        if user:
            print "Don't check for spam, because '%s' is a known local address for user '%s'" % (address,user.getName())
            return 0

        # get the receiver name. if it is a local receiver that doesn't want to have
        # spam checks, do so now
        for receiver in self.message['TO']:
            address = receiver
            if address[0] == '<' or address[-1] == '>':
                address = address[1:-1]

            print "checking receiver '%s'" % address

            # if the receiver is a local user
            user = ShicksUser.GetUserByAddress(address)
            if user:
                return user.getSpamCheck()
        return 1

    def RequiresPlaintextOnly(self):
        # get the receiver name. if it is a local receiver that doesn't want to have
        # spam checks, do so now
        for receiver in self.message['TO']:
            address = receiver
            if address[0] == '<' or address[-1] == '>':
                address = address[1:-1]

            # if the receiver is a local user
            user = ShicksUser.GetUserByAddress(address)
            if user:
                return user.getPlaintextOnly()
        return 0

    def CreateCopyOfCurrentMessage(self):

        # if required, check for incoming spam
        if self.RequiresSpamCheck():
            self.message["MSG"] = checkSpamMail(self.message["MSG"])

        if self.RequiresPlaintextOnly():
            self.message["MSG"] = plaintextOnly(self.message["MSG"])

        self.message["MSG"].insert(0,"Received: from %s (%s) by Shicks! Version %s on %s at %s\r\n"\
                                   % (self.sender_host, self.ip_address_string,
                                      CURRENT_VERSION, socket.gethostname(), time.ctime()))
        
        self.message["MSG"] = "".join(self.message["MSG"])
        self.message["MSGLEN"] = len(self.message["MSG"])       
        unique_filename = self.CreateUniqueMessageFilename()
        self.message["BODY"] = unique_filename
        file = open(unique_filename,"wb")
        file.write(self.message["MSG"])
        file.close()

    def StoreMessage(self,senderid,msgfilename,msglen,external):
        msg = ShicksMessageList.CreateEmptyMessage()
        msg.setUserId( senderid )
        msg.setFilename( msgfilename )
        msg.setMessageSize( msglen )
        msg.setDeleted( external )
        msg.setMessageType( external )
        msg.saveChanges()

    def SendMessage(self):
        "Called when the message has been received completely. See comment for StartReceivingData()"

        try:
            self.CreateCopyOfCurrentMessage()
        except:
            traceback.print_exc()
        
        self.receiving_data = 0

        # check if the mail can be processed. This is true, if
        # a) the sender is an internal address
        # b) all receivers have an internal addres
        
        can_process_message = 0

        # step 1: test if this is a message from a local user, or from an outside user        
        sender = self.GetUserForAddress(self.message['FROM'])
        sender_id = 0
        if sender:
            sender_id = sender.getId()

        self.message["sender-id"] = sender_id

        receivers = []

        if not sender and self.is_local_connection:
            sender = self.GetUserForAddress(ADMIN_ACCOUNT)
            sender_id = sender.getId()

        if sender:

            # insert message to db
            self.StoreMessage(sender_id,self.message["BODY"],self.message["MSGLEN"],1)
            
            # it is, so process locally
            for receiver in self.message['TO']:
                recuser = self.GetUserForAddress(receiver)
                if recuser:
                    # it is a local receiver, send immediately
                    self.SendCurrentMessageToLocalUser(recuser)
                else:
                    # it is a remote receiver, forward to other SMTP host.
                    receivers.append(receiver[1:-1])
                    
            # finished; check if mails have to dispatched externally
            if len(receivers):
                # try sending externally. if not successfull, fail

                self.SendCurrentMessageToRemoteUsers(sender_id,receivers)

            # done! 
        else:
            # it is a message from somebody outside
            for receiver in self.message['TO']:
                recuser = self.GetUserForAddress(receiver)
                if recuser:
                    # it is a local receiver, send immediately
                    self.SendCurrentMessageToLocalUser(recuser)
                else:
                    # it is a remote receiver, forward to other SMTP host.
                    receivers.append(receiver[1:-1])

            if len(receivers):
                print "Sorry, cannot dispatch to remote receivers '%s'" % str(receivers)
                msgbody = "A message from %s could not be sent to %s.\r\n" % (self.message['FROM'],",".join(receivers))
                msgbody += "The title of the message is '%s'." % self.GetCurrentSubject()
                self.SendAdminMail(msgbody)

        print "OK, sending OK"
        self.SendResponse('OK_DATA_RECV')

        # bugfix in 0.4 : reset messages
        self.ResetMessage()

    def CanSendMessage(self):
        global STRICT_RELAY_TEST

        # step 1: test if this is a message from a local user, or from an outside user        
        sender = self.GetUserForAddress(self.message['FROM'])

        # if the sender is a local mailbox, the message can be sent
        if sender:
            return 1

        # it is a message from somebody outside
        can_process_message, remote_receivers = 1, 0
        
        for receiver in self.message['TO']:
            recuser = self.GetUserForAddress(receiver)

            if STRICT_RELAY_TEST:
                # if strict testing is enabled, ALL receivers must be local
                if not recuser:
                    return 0

                continue

            elif recuser:
                # if strict testing is disabled, at least one reciever must be local
                return 1

            remote_receivers = 1

        if remote_receivers:
            can_process_message = 0
            assert( not STRICT_RELAY_TEST )

        return can_process_message

    def SendCurrentMessageToLocalUser(self,user):
        userid = user.getId()
       
        self.StoreMessage(userid,self.message["BODY"],self.message["MSGLEN"],0)

        remote_receivers = []

        # send to all forwarder entries
        for forwardentry in user.getForwarders():

            # handle recursive mail gracefully            
            if forwardentry[0] == userid:
                print "Warning, do not forward message to self!"

            # forward to internal user            
            elif forwardentry[0]:
                self.StoreMessage(forwardentry[0],self.message["BODY"],self.message["MSGLEN"],0)
            else:
                remote_receivers.append(forwardentry[1])
                
        if len(remote_receivers):
            self.SendCurrentMessageToRemoteUsers(userid,remote_receivers)
        
        return 1

    def CanSendMessageUsingServer(self, senderid, fromaddr, addrlist, messagebody, servername):
        print "trying to send message from %s to %s using server %s " % (fromaddr, ",".join(addrlist), servername)
        try:
            server = smtplib.SMTP(servername[1])
            #server.set_debuglevel(1)
            not_received = server.sendmail(fromaddr, addrlist, messagebody)
            server.quit()
            # Quote from the documentation: "If this method does not throw an exception, it returns a
            # dictionary, with one entry for each recipient that was refused. Each entry contains a
            # tuple of the SMTP error code and the accompanying error message sent by the server." EOQ
            if not_received:
                print "not_received=%s" % str(not_received)
                addrlist = filter(lambda x, keys=not_received.keys(): x not in keys, addrlist)
            else:
                addrlist = []
        except:
            temp = cStringIO.StringIO()
            traceback.print_exc(file=temp)
            traceback.print_exc()
            msg = temp.getvalue()
            temp.close()
            
            self.SendAdminMail("Exception caught while sending message titled\r\n\r\n'"+self.GetCurrentSubject()+\
                               "'\r\n\r\nto "+str(addrlist)+" using server '"+servername[1]+"'\r\n\r\n"+msg, senderid)

        return addrlist            

    def SendMailToUser(self, senderid, fromaddr, toaddrlist, messagebody):

        # first, for each target address, creates
        address_domain_assoc = {}
        for address in toaddrlist:
            domain = domainOfAddress(address)
            try:
                address_domain_assoc[domain].append( address )
            except:
                address_domain_assoc[domain] = [address]

        # now, send mail for all domains separately
        for domain in address_domain_assoc.keys():
            servers = mxlookup.MxRecordLookup(domain)
            addrlist = address_domain_assoc[domain]
            print "Domain '%s' has MX record %s" % (domain, str(servers))
            if servers is not None:
                if RELAY_SMTP_SERVER_NAME:
                    servers.append( (0, RELAY_SMTP_SERVER_NAME) ) # as a last resort
                for server in servers:
                    addrlist = self.CanSendMessageUsingServer(senderid, fromaddr, addrlist, messagebody, server)
                    print "addrlist is now %s" % str(addrlist)
                    if not len(addrlist): break
                # this message is not needed, because CanSendMessageUsingServer() will normally already report error messages.
                if addrlist and not servers:
                    self.SendAdminMail("Unable to send message with subject\r\n\r\n'"+self.GetCurrentSubject()+\
                                       "'\r\n\r\nto "+str(addrlist)+"\r\n", senderid)
            else:
                self.SendAdminMail("Unable to send message with subject\r\n\r\n'"+self.GetCurrentSubject()+\
                                   "'\r\n\r\nbecause unable to lookup MX record for domain '"+domain+"'\r\n", senderid)
                
    def SendCurrentMessageToRemoteUsers(self,senderid,receivers):
        try:
            global RELAY_SMTP_SERVER_NAME, USE_RELAY_SMTP_SERVER
            if not RELAY_SMTP_SERVER_NAME and USE_RELAY_SMTP_SERVER:
                print "Sorry, no relay server configured."
                self.SendAdminMail("Sorry, no relay server configured. This means that the mail has not been forwarded to these addresses: %s" % ",".join(receivers), self.message["sender-id"])
                return

            print "sending message to remote users '%s'" % str(receivers)
            self.SendMailToUser(senderid, self.message["FROM"][1:-1], receivers, self.message['MSG'])
        except:
            temp = cStringIO.StringIO()
            traceback.print_exc(file=temp)
            traceback.print_exc()
            msg = temp.getvalue()
            temp.close()
            
            self.SendAdminMail("Exception caught while sending message titled\r\n\r\n'"+self.GetCurrentSubject()+\
                               "'\r\n\r\nto "+str(receivers)+"\r\nTraceback follows:\r\n"+msg, senderid)

    def GetCurrentSubject(self):
        messagebody = self.message['MSG']
        if type(messagebody) == type([]):
            for line in messagebody:
                if line[:8].lower() == 'subject:':
                    return line[8:].strip()
        else:
            index = messagebody[:1000].lower().find("subject:")
            if index >= 0:
                line = messagebody[index+8:index+200]
                index = line.find("\n")
                if index >= 0:
                    line = line[:index]
                line = line.strip()
                return line
                
        return '<none>'

    def SendAdminMail(self,msgbody,receiver_id=None):
        global ADMIN_ACCOUNT
        

        admin = ShicksUser.GetUserByAddress(ADMIN_ACCOUNT)
        if not admin:
            print "Warning, unable to send admin mail, because no postmaster is configured."
            return None

        if receiver_id is None:
            receiver_id = admin.getId()
            
        msg = "From: Shicks! Mail-Administrator\r\nTo: %s\r\n" % admin.getName()
        msg += "Subject: Warning from Shicks! Mail-Administrator\r\nDate:%s\r\n" % time.asctime(time.localtime())
        msg += "Importance: High\r\n"
        msg += "\r\n"
        msg += msgbody

        unique_filename = self.CreateUniqueMessageFilename()
        file = open(unique_filename,"wb")
        file.write(msg)
        file.close()
        
        self.StoreMessage(receiver_id,unique_filename,len(msg),0)
        return 1
        
    def GetUserForAddress(self,address):
        "Returns the user object for a given email address, or None if the address is unknown"
        if address[0] == '<' or address[-1] == '>':
            address = address[1:-1]

        user = ShicksUser.GetUserByAddress(address)
        if user:
            return user

        # map unknown addresses only, if CATCH_ALL is enabled.
        # it is disabled per default.
        if (domainOfAddress(address).lower() == DEFAULT_LOCAL_DOMAIN.lower()):
            if CATCH_ALL:
                user = ShicksUser.GetUserByAddress(UNKNOWN_ACCOUNT)
                if user:
                    print "Because of CATCH-ALL, %s is mapped to %s" % (address, UNKNOWN_ACCOUNT)
                    return user

    def GetExpandedName(self,name):
        
        if name[0] == '<' and name[-1] == '>':
            name = name[1:-1]
        
        # step 1: look up address directly"
        user = ShicksUser.GetUserByAddress(name)
        if not user:
            user = ShicksUser.GetUserByName(name)
        if not user:
            user = ShicksUser.GetUserByID(name)
            
        if user:
            return "<%s>" % name

    def IsReceiverAllowed(self,address):
        global ALLOW_UNKNOWN_RECEIVER_DEFAULT

        if address[0] == '<' or address[-1] == '>':
            address = address[1:-1]

        # New in Version 0.4: If the sender is a local address, it must
        # come from a local IP, otherwise access is denied
        if STRICT_RELAY_TEST and IP_ADDRESS_RANGE and not self.is_local_connection:
            user = ShicksUser.GetUserByAddress(address)
            if user:
                print "Target address '%s' is local." % address
                return 1
            else:
                fromstr = str(self.message["FROM"])
                if fromstr not in ('','<>'):
                    print "Attempt to send message from '%s' to '%s' denied." % (fromstr, address)
                    self.SendAdminMail("""Warning, an attempt has been made to send a message 
    from '%s' to '%s'. The access has been denied.""" % (str(self.message["FROM"]), address))
                return 0
        
        if addressInRange( address, DENY_HOSTS_LIST ):
            print "Address '%s' is on the deny-hosts list" % address
            return 0

        # if it is a local address, but unknown, and CATCH_ALL disabled, forbid sending
        if not CATCH_ALL and (domainOfAddress(address).lower() == DEFAULT_LOCAL_DOMAIN.lower()):
            user = ShicksUser.GetUserByAddress(address)
            if user:
                print "Target address '%s' is local." % address
            else:               
                print "Because CATCH-ALL is disabled, %s is being denied access" % address
                return 0

        return 1

    def IsSenderAllowed(self,address):

        if address[0] == '<' or address[-1] == '>':
            address = address[1:-1]

        # New in Version 0.4: If the sender is a local address, it must
        # come from a local IP, otherwise access is denied
        if STRICT_RELAY_TEST and IP_ADDRESS_RANGE and not self.is_local_connection:
            user = ShicksUser.GetUserByAddress(address)
            if user:
                print "Error, sender '%s' has a local e-mail address, but comes from a remote IP '%s'" % (address, self.ip_address_string)
                self.SendAdminMail("""Warning, an attempt has been made to send a message 
from IP '%s' with the local e-mail address '%s'. The access has been denied.""" % (self.ip_address_string, address))
                return 0
            else:
                print "Sender '%s' is not a known local address, so use of remote IP "\
                      "'%s' is possible." % (address, self.ip_address_string)

        if addressInRange( address, DENY_HOSTS_LIST ):
            print "Address '%s' is on the deny-hosts list" % address
            return 0

        return 1
        
def StartSMTPServer():
    server = SMTPServer(('', 25), SMTPRequestHandler)
    print "SMTP server starting - Version %s on %s" % (CURRENT_VERSION,socket.gethostname())
    server.serve_forever()
    print "SMTP server returned"

if __name__ == '__main__':
    # read configuration
    TRACE_DIRECTORY = GetConfigSettings("Logfiles",(os.name == "nt") and "c:\\shicks\\log" or "/usr/share/shicks/log")
    try:
        os.makedirs(TRACE_DIRECTORY)
    except:
        pass
    ShicksTrace.TraceFileName = os.path.join(TRACE_DIRECTORY,"SMTPServer.log")
    ShicksTrace.Enable()

    StartSMTPServer()






